@@ -1,6 +1,6 @@
 /*
- * Copyright (C) 2015 - 2019, Stephan Mueller <smueller@chronox.de>
- * Copyright (C) 2019, Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2015 - 2020, Stephan Mueller <smueller@chronox.de>
+ * Copyright (C) 2020, Red Hat, Inc. All rights reserved.
  *
  * License: see LICENSE file in root directory
  *
@@ -56,6 +56,7 @@
 #include <dlfcn.h>
 #include <libgen.h>
 #include <limits.h>
+#include <inttypes.h>
 
 #include <kcapi.h>
 
@@ -73,7 +74,7 @@ struct hash_name {
 struct hash_key {
 	const char *checkdir;
 	const uint8_t *data;
-	uint32_t len;
+	off_t len;
 };
 
 struct hash_params {
@@ -81,6 +82,7 @@ struct hash_params {
 	struct hash_key key;
 	uint32_t hashlen;
 	int bsd_style;
+	uint32_t newline;
 };
 
 static const struct hash_name NAMES_MD5[2] = {
@@ -152,6 +154,7 @@ static void usage(char *name, int fipscheck)
 	fprintf(stderr, "\t   --tag\t\tCreate a BSD-style checksum\n");
 	fprintf(stderr, "\t-d\t\t\tCheck directory for fipshmac; otherwise ignored\n");
 	fprintf(stderr, "\t-b, -P\t\t\tCompatibility hmaccalc options; ignored\n");
+	fprintf(stderr, "\t-z\t\t\tNUL line termination\n");
 	fprintf(stderr, "\t   --help\t\tPrint this help text\n");
 	fprintf(stderr, "\t-v --version\t\tShow version\n");
 }
@@ -166,7 +169,8 @@ static void version(char *name)
 	fprintf(stderr, "%s: %s\n", basename(name), version);
 }
 
-static int mmap_file(const char *filename, uint8_t **memory, uint32_t *size)
+static int mmap_file(const char *filename, uint8_t **memory, off_t *size,
+		     size_t *mapped, off_t offset)
 {
 	int fd = -1;
 	int ret = 0;
@@ -179,35 +183,42 @@ static int mmap_file(const char *filename, uint8_t **memory, uint32_t *size)
 		return -EIO;
 	}
 
-	ret = check_filetype(fd, &sb, filename);
-	if (ret)
-		goto out;
-
-	*memory = NULL;
-	*size = sb.st_size;
-
-	if (sb.st_size) {
-		*memory = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);
-		if (*memory == MAP_FAILED)
-		{
-			*memory = NULL;
-			fprintf(stderr, "Use of mmap failed\n");
-			ret = -ENOMEM;
+	if (*size) {
+		if ((*size - offset) < (off_t)*mapped )
+			*mapped = (size_t)(*size - offset);
+	} else {
+		ret = check_filetype(fd, &sb, filename);
+		if (ret)
+			goto out;
+		*size = sb.st_size;
+		if (*size <= (off_t)*mapped) {
+			*mapped = (size_t)*size;
+		if (*size == 0)
 			goto out;
 		}
 	}
+
+	*memory = mmap(NULL, *mapped, PROT_READ, MAP_PRIVATE | MAP_POPULATE, fd,
+		       offset);
+	if (*memory == MAP_FAILED) {
+		*memory = NULL;
+		ret = -errno;
+		goto out;
+	}
+	madvise(*memory, *mapped, MADV_SEQUENTIAL | MADV_WILLNEED);
+
 out:
 	close(fd);
 	return ret;
 }
 
-static int load_file(const char *filename, uint8_t **memory, uint32_t *size)
+static int load_file(const char *filename, uint8_t **memory, off_t *size)
 {
 	int fd = -1;
 	int ret = 0;
 	uint8_t *buffer = NULL;
 	uint32_t buffer_size = 4096;
-	size_t offset = 0;
+	off_t offset = 0;
 	ssize_t rdbytes;
 
 	fd = open(filename, O_RDONLY | O_CLOEXEC);
@@ -232,8 +243,8 @@ static int load_file(const char *filename, uint8_t **memory, uint32_t *size)
 			goto out;
 		}
 
-		offset += (size_t)rdbytes;
-		if (offset == buffer_size) {
+		offset += (off_t)rdbytes;
+		if (offset == (off_t)buffer_size) {
 			uint8_t *new_buffer;
 
 			if (buffer_size == UINT32_MAX) {
@@ -257,7 +268,7 @@ static int load_file(const char *filename, uint8_t **memory, uint32_t *size)
 	}
 
 	*memory = buffer;
-	*size = (uint32_t)offset;
+	*size = offset;
 
 	close(fd);
 	return 0;
@@ -272,35 +283,49 @@ out:
 static int hasher(struct kcapi_handle *handle, const struct hash_params *params,
 		  const char *filename, const char *comphash, uint32_t comphashlen,
 		  FILE *outfile)
-{	
+{
+	/* Mapping file in 16M segments */
+	size_t mapped = 16<<20;
+	off_t offset = 0, size = 0;
+	uint32_t hashlen = params->hashlen;
 	int ret = 0;
 	uint8_t *memblock = NULL;
 	uint8_t *memblock_p;
-	uint32_t size, left, hashlen = params->hashlen;
 	uint8_t md[64];
 
 	if (filename) {
-		ret = mmap_file(filename, &memblock, &size);
-		if (ret)
-			goto out;
-		/* Compute hash */
-		memblock_p = memblock;
-		left = size;
-		while (left) {
-			uint32_t todo = (left > INT_MAX) ? INT_MAX : left;
-
-			ret = kcapi_md_update(handle, memblock_p, todo);
-			if (ret < 0)
+		do {
+			ret = mmap_file(filename, &memblock, &size, &mapped,
+					offset);
+			if (ret) {
+				fprintf(stderr,
+					"Use of mmap failed mapping %zu bytes at offset %" PRId64 " of file %s (%d)\n",
+					mapped, (int64_t)offset, filename, ret);
 				goto out;
-			left -= todo;
-			memblock_p += todo;
-		}
+			}
+			/* Compute hash */
+			memblock_p = memblock;
+			size_t left = mapped;
+			do {
+				uint32_t todo = (left > INT_MAX) ?
+						 INT_MAX : (uint32_t)left;
+
+				ret = kcapi_md_update(handle, memblock_p, todo);
+				if (ret < 0)
+					goto out;
+				left -= todo;
+				memblock_p += todo;
+			} while (left);
+			munmap(memblock, mapped);
+			offset = offset + mapped;
+		} while (offset ^ size);
 	} else {
 		uint8_t tmpbuf[TMPBUFLEN] __aligned(KCAPI_APP_ALIGN);
-		size_t bufsize;
+		uint32_t bufsize;
 
 		while ((bufsize =
-			fread(tmpbuf, sizeof(uint8_t), TMPBUFLEN, stdin))) {
+			(uint32_t)fread(tmpbuf, sizeof(uint8_t), TMPBUFLEN,
+					stdin))) {
 
 			ret = kcapi_md_update(handle, tmpbuf, bufsize);
 			if (ret < 0)
@@ -333,14 +358,17 @@ static int hasher(struct kcapi_handle *handle, const struct hash_params *params,
 				ret = 0;
 		} else {
 			if (outfile == NULL) { /* only print hash (hmaccalc -S) */
-				bin2print(md, hashlen, NULL, stdout, 1);
+				bin2print(md, hashlen, NULL, stdout,
+					  params->newline);
 			} else if (params->bsd_style) {
-				fprintf(outfile, "%s (%s) = ", params->name.bsdname,
+				fprintf(outfile, "%s (%s) = ",
+					params->name.bsdname,
 					filename ? filename : "-");
-				bin2print(md, hashlen, NULL, outfile, 1);
+				bin2print(md, hashlen, NULL, outfile,
+					  params->newline);
 			} else {
 				bin2print(md, hashlen, filename ? filename : "-",
-					  outfile, 1);
+					  outfile, params->newline);
 			}
 			ret = 0;
 		}
@@ -351,7 +379,7 @@ static int hasher(struct kcapi_handle *handle, const struct hash_params *params,
 
 out:
 	if (memblock)
-		munmap(memblock, size);
+		munmap(memblock, mapped);
 
 	return ret;
 }
@@ -386,10 +414,16 @@ static char *paste(char *dst, const char *src, size_t size)
 static char *get_hmac_file(const char *filename, const char *checkdir)
 {
 	size_t i, filelen, pathlen, namelen, basenamestart = 0;
-	size_t prefixlen = strlen(CHECK_PREFIX);
+	const char *check_prefix = CHECK_PREFIX;
+	size_t prefixlen = strlen(check_prefix);
 	size_t suffixlen = strlen(CHECK_SUFFIX);
 	char *cursor, *checkfile = NULL;
 
+	if (prefixlen == 0 && checkdir == NULL) {
+		check_prefix = ".";
+		prefixlen = 1;
+	}
+
 	filelen = strlen(filename);
 	if (filelen > 4096) {
 		fprintf(stderr, "File too long\n");
@@ -415,7 +449,7 @@ static char *get_hmac_file(const char *filename, const char *checkdir)
 	} else if (pathlen > 0)
 		cursor = paste(cursor, filename, pathlen);
 
-	cursor = paste(cursor, CHECK_PREFIX, prefixlen);
+	cursor = paste(cursor, check_prefix, prefixlen);
 	cursor = paste(cursor, filename + basenamestart, namelen);
 	cursor = paste(cursor, "."CHECK_SUFFIX, 1 + suffixlen);
 	strncpy(cursor, "\0", 1);
@@ -439,7 +473,8 @@ static int hash_files(const struct hash_params *params,
 		return -EFAULT;
 	}
 	if (params->key.data) {
-		ret = kcapi_md_setkey(handle, params->key.data, params->key.len);
+		ret = kcapi_md_setkey(handle, params->key.data,
+				      (uint32_t)params->key.len);
 		if (ret) {
 			fprintf(stderr, "Setting HMAC key for %s failed (%d)\n",
 				hashname, ret);
@@ -512,7 +547,8 @@ static int process_checkfile(const struct hash_params *params,
 		return -EFAULT;
 	}
 	if (params->key.data) {
-		ret = kcapi_md_setkey(handle, params->key.data, params->key.len);
+		ret = kcapi_md_setkey(handle, params->key.data,
+				      (uint32_t)params->key.len);
 		if (ret) {
 			fprintf(stderr, "Setting HMAC key for %s failed (%d)\n",
 				hashname, ret);
@@ -532,7 +568,7 @@ static int process_checkfile(const struct hash_params *params,
 		char *filename = NULL;   // parsed file name
 		char *hexhash = NULL;    // parsed hex value of hash
 		uint32_t hexhashlen = 0; // length of hash hex value
-		uint32_t linelen = strlen(buf);
+		uint32_t linelen = (uint32_t)strlen(buf);
 		uint32_t i;
 		uint32_t bsd_style = 0; // >0 if --tag formatted style
 
@@ -669,7 +705,7 @@ static int fipscheck_self(const struct hash_params *params_bin,
 #define BUFSIZE 4096
 	char selfname[BUFSIZE];
 	char *names[] = { selfname };
-	int32_t selfnamesize = 0;
+	ssize_t selfnamesize = 0;
 	Dl_info info;
 	void *dl = NULL, *sym;
 
@@ -694,7 +730,7 @@ static int fipscheck_self(const struct hash_params *params_bin,
 			}
 		}
 
-		n = fread((void *)fipsflag, 1, 1, fipsfile);
+		n = (uint32_t)fread((void *)fipsflag, 1, 1, fipsfile);
 		fclose(fipsfile);
 		if (n != 1) {
 			fprintf(stderr, "Cannot read FIPS flag\n");
@@ -797,11 +833,14 @@ int main(int argc, char *argv[])
 		.key = { NULL, NULL, 0 },
 		.hashlen = 0,
 		.bsd_style = 0,
+		.newline = 1,
 	};
 	const struct hash_params *params_self;
 	char *basec = NULL;
 	const char *basen = NULL;
 	int ret = -EFAULT;
+	/* File memory-mapping size limit set at 64MB in 32bit and 1GB in 64bit virtual memory space */
+	size_t mapped = (sizeof(void*) == 4) ? 64<<20 : 1<<30;
 
 	char *checkfile = NULL;
 	const char *targetfile = NULL;
@@ -821,7 +860,7 @@ int main(int argc, char *argv[])
 		{0, 0, 0, 0}
 	};
 
-	static const char *opts_short = "c:uh:t:SLqk:K:vbd:P";
+	static const char *opts_short = "c:uh:t:SLqk:K:vbd:Pz";
 	static const struct option opts[] = {
 		{"help", 0, 0, 0},
 		{"tag", 0, 0, 0},
@@ -924,6 +963,7 @@ int main(int argc, char *argv[])
 		params_self = &PARAMS_SELF_HMACCALC;
 	} else {
 		fprintf(stderr, "Unknown invocation name: %s\n", basen);
+		usage(argv[0], fipscheck);
 		ret = 1;
 		goto out;
 	}
@@ -962,7 +1002,7 @@ int main(int argc, char *argv[])
 			case 'u':
 				if (hmackey_alloc) {
 					kcapi_memset_secure(hmackey_alloc, 0,
-					                    params.key.len);
+						(uint32_t)params.key.len);
 					free(hmackey_alloc);
 					hmackey_alloc = NULL;
 				} else if (hmackey_mmap) {
@@ -1016,21 +1056,25 @@ int main(int argc, char *argv[])
 			case 'k':
 				if (hmackey_alloc) {
 					kcapi_memset_secure(hmackey_alloc, 0,
-					                    params.key.len);
+						(uint32_t)params.key.len);
 					free(hmackey_alloc);
 					hmackey_alloc = NULL;
 				} else if (hmackey_mmap) {
 					munmap(hmackey_mmap, params.key.len);
 					hmackey_mmap = NULL;
 				}
-				ret = mmap_file(optarg, &hmackey_mmap, &params.key.len);
+				params.key.len = 0;
+				ret = mmap_file(optarg, &hmackey_mmap,
+						&params.key.len,
+						&mapped, 0);
 				if (!ret) {
 					params.key.data = hmackey_mmap;
 					hmac = 1;
 					break;
 				}
 				/* fallback to normal file I/O: */
-				ret = load_file(optarg, &hmackey_alloc, &params.key.len);
+				ret = load_file(optarg, &hmackey_alloc,
+						&params.key.len);
 				if (ret) {
 					ret = 1;
 					goto out;
@@ -1041,7 +1085,7 @@ int main(int argc, char *argv[])
 			case 'K':
 				if (hmackey_alloc) {
 					kcapi_memset_secure(hmackey_alloc, 0,
-					                    params.key.len);
+						(uint32_t)params.key.len);
 					free(hmackey_alloc);
 					hmackey_alloc = NULL;
 				} else if (hmackey_mmap) {
@@ -1069,6 +1113,9 @@ int main(int argc, char *argv[])
 			case 'P':
 				/* Compatibility options, just ignore */
 				break;
+			case 'z':
+				params.newline = 2;
+				break;
 			default:
 				usage(argv[0], fipscheck);
 				ret = 1;
@@ -1128,7 +1175,8 @@ int main(int argc, char *argv[])
 	}
 
 	if (!checkfile)
-		ret = hash_files(&params, argv + optind, (argc - optind),
+		ret = hash_files(&params, argv + optind,
+				 (uint32_t)(argc - optind),
 		                 fipshmac, checkdir, 0);
 	else if (optind == argc)
 		ret = process_checkfile(&params, checkfile, targetfile, loglevel);
@@ -1144,7 +1192,7 @@ out:
 	if (checkfile)
 		free(checkfile);
 	if (hmackey_alloc) {
-		kcapi_memset_secure(hmackey_alloc, 0, params.key.len);
+		kcapi_memset_secure(hmackey_alloc, 0, (uint32_t)params.key.len);
 		free(hmackey_alloc);
 	} else if (hmackey_mmap) {
 		munmap(hmackey_mmap, params.key.len);
